//
// Configuration.C -- News configuration class
//
// Copyright 2003-2004 Michael Sweet
// Copyright 2002 Greg Ercolano
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public Licensse as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include "everything.H"
#include <stdarg.h>
#include <syslog.h>


// Initialize default configuration values...
Configuration::Configuration()
{
   char	name[1024];			// Hostname string


   // Set defaults...
   HostnameLookups(0);

   Listen(119);

   log = 0;
   LogFile("stderr");

   LogLevel(L_INFO);

   MaxClients(0);

   gethostname(name, sizeof(name));
   NewsHostname(name);

   SendMail(SENDMAIL);

   SpamFilter("");

   SpoolDir(SPOOL_DIR);

   Timeout(12 * 3600);

   User("news");
}


// Listen on a specific address and port...
void Configuration::Listen(const char *l)
{
    char		hostname[256];	// Hostname or IP
    char		portname[256];	// Port number or name
    char		*ptr;		// Pointer into port
    struct hostent	*host;		// Host address
    struct servent	*port;		// Service data
    long		p;		// Port number


    // Initialize the listen address to "localhost:nntp"...
    memset(&listen, 0, sizeof(struct sockaddr_in));
    listen.sin_family      = AF_INET;
    listen.sin_addr.s_addr = htonl(0x7f000001);
    listen.sin_port        = htons(119);

    // Try to grab a hostname and port number...
    switch (sscanf(l, "%255[^:]:%255s", hostname, portname))
    {
	case 1 :
            if (strchr(hostname, '.') == NULL)
	    {
	       /*
		* Hostname is a port number...
		*/

		strcpy(portname, hostname);
		strcpy(hostname, "*");
	    }
            else
        	portname[0] = '\0';
            break;

	case 2 :
            break;

	default :
	    fprintf(stderr, "news: Unable to decode address \"%s\"!\n", l);
            return;
    }

   /*
    * Decode the hostname and port number as needed...
    */

    if (hostname[0] && strcmp(hostname, "*"))
    {
	if ((host = gethostbyname(hostname)) == NULL)
	{
	    fprintf(stderr, "newsd: gethostbyname(\"%s\") failed - %s!\n",
	            hostname, hstrerror(h_errno));
	    fputs("newsd: Using address 127.0.0.1 (localhost)\n", stderr);
	}
	else if (host->h_length != 4 || host->h_addrtype != AF_INET)
	{
	    fprintf(stderr, "newsd: gethostbyname(\"%s\") did not return an IPv4 address!\n",
        	    hostname);
	    fputs("newsd: Using address 127.0.0.1 (localhost)\n", stderr);
	}
	else
	    memcpy(&(listen.sin_addr), host->h_addr, 4);
    }
    else if (!strcmp(hostname, "*"))
	listen.sin_addr.s_addr = INADDR_ANY;

    if (portname[0] != '\0')
    {
	if (isdigit(portname[0]))
	{
	    p = strtol(portname, &ptr, 10);
	    if (p <= 0 || *ptr)
	    {
	        fprintf(stderr, "newsd: Bad port number \"%s\"!\n", portname);
        	fputs("newsd: Using port 119 (nntp)\n", stderr);
            }
	    else
		listen.sin_port = htons(p);
        }
	else
	{
	    if ((port = getservbyname(portname, "tcp")) == NULL)
	    {
        	fprintf(stderr, "newsd: getservbyname(\"%s\", \"tcp\") failed - %s!\n",
	        	portname, strerror(errno));
        	fputs("newsd: Using port 119 (nntp)\n", stderr);
	    }
	    else
        	listen.sin_port = port->s_port;
	}
    }
}


// Listen on a port...
void Configuration::Listen(int p)
{
    // Listen on the "any" address with the specified port; needs to be
    // updated for IPv6 at some point...
    memset(&listen, 0, sizeof(struct sockaddr_in));
    listen.sin_family = AF_INET;
    listen.sin_port   = htons(p);
}


// Load configuration values from a file...
void Configuration::Load(const char *conffile)
{
    FILE	*fp;			// Config file
    int		linenum;		// Line number
    char	line[1024],		// Line from config file
		*ptr,			// Pointer into line
		*name,			// Config directive
		*value;			// Config value
    long	lvalue;			// Integer value


    if ((fp = fopen(conffile, "r")) == NULL)
    {
        fprintf(stderr, "newsd: Unable to open configuration file \"%s\" - %s!\n",
	        conffile, strerror(errno));
        return;
    }

    linenum = 0;

    while (fgets(line, sizeof(line), fp))
    {
        // Increment line counter...
        linenum ++;

        // Strip comments...
	if ((ptr = strchr(line, '#')) != NULL)
	    *ptr = '\0';

        // Strip trailing whitespace...
	for (ptr = line + strlen(line) - 1; ptr >= line && isspace(*ptr); ptr --)
            *ptr = '\0';

	// Ignore blank lines...
	if (!line[0])
	    continue;

        // Find directive name...
	for (name = line; isspace(*name); name ++);

        // Separate directive from value...
	for (value = name + 1; !isspace(*value) && *value; value ++);

	while (isspace(*value))
	  *value++ = '\0';

	if (!*value)
	{
	    fprintf(stderr, "newsd: Missing value on line %d of \"%s\"!\n", linenum,
	            conffile);
	    continue;
	}

        // Decode configuration directive...
	if (!strcasecmp(name, "HostnameLookups"))
	{
	    if (!strcasecmp(value, "off") || !strcasecmp(value, "no"))
	    {
	        HostnameLookups(0);
	    }
	    else if (!strcasecmp(value, "on") || !strcasecmp(value, "yes"))
	    {
	        HostnameLookups(1);
	    }
	    else if (!strcasecmp(value, "double"))
	    {
	        HostnameLookups(2);
	    }
	    else
		fprintf(stderr, "newsd: Bad HostnameLookups value \"%s\" on line %d of \"%s\"!\n",
		        value, linenum, conffile);
	}
	else if (!strcasecmp(name, "Listen") || !strcasecmp(name, "Port"))
	{
	    Listen(value);
	}
	else if (!strcasecmp(name, "LogFile"))
	{
	    LogFile(value);
	}
	else if (!strcasecmp(name, "LogLevel"))
	{
	    if (!strcasecmp(value, "error"))
	        LogLevel(L_ERROR);
	    else if (!strcasecmp(value, "info"))
	        LogLevel(L_INFO);
	    else if (!strcasecmp(value, "debug"))
	        LogLevel(L_DEBUG);
	    else
		fprintf(stderr, "newsd: Bad LogLevel value \"%s\" on line %d of \"%s\"!\n",
		        value, linenum, conffile);
	}
	else if (!strcasecmp(name, "MaxClients"))
	{
	    lvalue = strtol(value, &ptr, 10);

	    if (lvalue < 0 || *ptr)
	        fprintf(stderr, "newsd: Bad MaxClients value \"%s\" on line %d of \"%s\"!\n",
		        value, linenum, conffile);
	    else
	        MaxClients(lvalue);
	}
	else if (!strcasecmp(name, "NewsHostname"))
	{
	    NewsHostname(value);
	}
	else if (!strcasecmp(name, "SendMail"))
	{
	    SendMail(value);
	}
	else if (!strcasecmp(name, "SpamFilter"))
	{
	    SpamFilter(value);
	}
	else if (!strcasecmp(name, "SpoolDir"))
	{
	    SpoolDir(value);
	}
	else if (!strcasecmp(name, "Timeout"))
	{
	    lvalue = strtol(value, &ptr, 10);

	    if (lvalue < 0 || *ptr)
	        fprintf(stderr, "newsd: Bad Timeout value \"%s\" on line %d of \"%s\"!\n",
		        value, linenum, conffile);
	    else
	        Timeout(lvalue);
	}
	else if (!strcasecmp(name, "User"))
	{
	    User(value);

            if (bad_user)
	        fprintf(stderr, "newsd: Bad User value \"%s\" on line %s of \"%s\"!\n",
		        value, linenum, conffile);
	}
    }

    // Close the config file and return...
    fclose(fp);
}


// Set the log file...
void Configuration::LogFile(const char *f)
{
    if (log)
    {
        if (log != stderr)
            fclose(log);

	log = 0;
    }

    logfile = f;

    if (logfile == "syslog")
        openlog("newsd", LOG_PID, LOG_NEWS);
    else if (logfile == "stderr")
        log = stderr;
    else
    {
        log = fopen(logfile.c_str(), "a");
	if (!log)
	{
	    fprintf(stderr, "newsd: Unable to open log file \"%s\" - %s.\n",
	            logfile.c_str(), strerror(errno));
            log = stderr;
	}
	else
	    setbuf(log, NULL);

        fchmod(fileno(log), 0600);
	fchown(fileno(log), uid, gid);
    }
}


// Log a message...
void Configuration::LogMessage(int l, const char *m, ...)
{
    // Only log messages as needed...
    if (l > loglevel)
        return;

    // Format the message...
    va_list      ap;			// Argument list pointer
    char         buffer[1024];		// Message buffer
    unsigned int bytes;			// Size of message


    va_start(ap, m);
    bytes = vsnprintf(buffer, sizeof(buffer), m, ap);
    va_end(ap);

    // Send it to the log file or syslog...
    if (log)
    {
	time_t		secs;		// Current UNIX time
	struct tm	*date;		// Current date/time
	char		datestr[1024];	// Date/time string


	time(&secs);
	date = localtime(&secs);
	strftime(datestr, sizeof(datestr), "%c", date);

	fprintf(log, "%s newsd[%d]: %s%s", datestr, getpid(), buffer,
	        (bytes >= sizeof(buffer) ||
		 buffer[bytes - 1] != '\n') ? "\n" : "");

	fflush(log);
    }
    else
        syslog(loglevel == L_ERROR ? LOG_ERR :
	           loglevel == L_INFO ? LOG_INFO : LOG_DEBUG, "%s", buffer);
}

void Configuration::LogSelf(int loglevel)
{
    LogMessage(loglevel, "HostnameLookups %s",
                      HostnameLookups() == 0 ? "off" :
	                  HostnameLookups() == 1 ? "on" : " double");
    struct sockaddr_in *addr = Listen();
    unsigned ipaddr = ntohl(addr->sin_addr.s_addr);
    LogMessage(loglevel, "Listen %u.%u.%u.%u:%d",
        	      (ipaddr >> 24) & 255, (ipaddr >> 16) & 255,
		      (ipaddr >> 8) & 255, ipaddr & 255,
		      ntohs(addr->sin_port));
    LogMessage(loglevel, "LogFile %s", LogFile());
    LogMessage(loglevel, "LogLevel %s",
        	      LogLevel() == L_ERROR ? "error" :
	        	  LogLevel() == L_INFO ? "info" : "debug");
    LogMessage(loglevel, "MaxClients %u", MaxClients());
    LogMessage(loglevel, "NewsHostname %s", NewsHostname());
    LogMessage(loglevel, "SendMail %s", SendMail());
    LogMessage(loglevel, "SpamFilter %s", SpamFilter());
    LogMessage(loglevel, "SpoolDir %s", SpoolDir());
    LogMessage(loglevel, "Timeout %u", Timeout());
    LogMessage(loglevel, "User %s", User());
}


int
Configuration::lookup_user(const char *name)
{
    struct passwd *pw = getpwnam(name);

    if (pw == NULL )
    {
        uid      = UINT_MAX;
	gid      = UINT_MAX;
	bad_user = true;

        return(-1);
    }

    uid      = pw->pw_uid;
    gid      = pw->pw_gid;
    bad_user = false;
    endpwent();

    return(0);
}

